/*
	Shag: Hair plug-in for 3D Studio MAX
	Copyright (c) 1997-2000, Dimension DESIGN Animation Group

	File: ShagHairExport.h
	Created by: Ivan Kolev

	Export data structures
*/

#ifndef _SHAGHAIREXPORT_H_
#define _SHAGHAIREXPORT_H_


#define SHAGHAIR_CLASS_ID		Class_ID( 0x6d707903, 0x614216f )
#define SHAGFUR_CLASS_ID		Class_ID( 0x790b49f8, 0x74687d62 )
#define SHAGRENDER_CLASS_ID		Class_ID( 0x2b22403b, 0x168608d6 )
#define SHAGVIEW_CLASS_ID		Class_ID( 0x54d9504e, 0x48822ea3 )

#define MODELHAIR_CLASS_ID		Class_ID( 0x6af3cd97, 0x99c3b9db )
#define VECTOROBJ_CLASS_ID		Class_ID( 0x265057db, 0x378a63fe )

#define HAIRSHADER_CLASS_ID		Class_ID( 0x4eac6307, 0x1ba16fc3 )
#define HAIRSHADOW_CLASS_ID		Class_ID( 0x61ff0f35, 0x4a00c2f )

// these are the various types to which a ShagView object can be converted via
// Object::ConvertToType

// this returns the ShagExportData structure described below. It contains all
// the info about the final hair object. Be careful, it can get very large
#define SHAGEXPORTDATA_CLASS_ID		Class_ID( 0x712c5194, 0x5e9448e0 )

// this converts the vertex strands into a mesh suitable for cloth simulation
// each hair segment is turned into a degenerated face of the kind (v0,v0,v1),
// where v0 and v1 are the ends of the segment. so far, only SimCloth by Chaos
// Group is proven to work well with such objects
#define SHAGCLOTH_CLASS_ID			Class_ID( 0x67a86f1d, 0x468804d6 )

// this returns a MAX spline object consisting of the vertex strands only
#define SHAGVERTEXSPLINE_CLASS_ID	Class_ID( 0x78b2571b, 0x713540d3 )

// this returns a MAX mesh object with the following structure:
// numVerts, numFaces - equal to the number of strands in the final hair object
// verts[i]           - contains the coordinates of the root of the i-th strand of
//                      the final hair object, in the object space of the emitter
// faces[i].v         - contains {i,i,i}
// faces[i].smGroup   - contains the id of the emitter face to which the 
//                      i-th root belongs to
// tVerts[i]          - contains the barycentric coordinates of the i-th root
//                      relative to the emitter face it belongs to
#define SHAGSTRANDROOTS_CLASS_ID	Class_ID( 0x6ae6336e, 0x66bc16d3 )

// and of course, don't forget splineShapeClassID and triObjectClassID! :-)
// these are the standard MAX spline and mesh objects, which will contain the
// final state of the hair and can get very, very large


// the following defines the correspondance of Shag's mapping channels
// to MAX's mapping channels. Shag always uses the channel ordering of MAX R3,
// in which channel 0 is the vertex color and channels 1 and above (if any) are
// the standard mapping channels. for R2.x channels 0 and 1 must be swapped
#if GET_MAX_RELEASE(VERSION_3DSMAX) >= 3000		// R3+
#define	MAX_MAPCHANNELS			MAX_MESHMAPS
#define	MAPCHANNEL_VERTCOL		0
#define	MAPCHANNEL_FIRST		1
#define MapChanHair2Max(iChannel)		(iChannel)
#else											// R2
#define MAX_MAPCHANNELS			2
#define	MAPCHANNEL_VERTCOL		1
#define	MAPCHANNEL_FIRST		0
#define MapChanHair2Max(iChannel)		( 1 - (iChannel) )
#endif

// Shag UV mapping types (selected with the UV Mapping controls in the Shading rollout)
enum MapType {
	MAP_VERTEXCOLOR = -3,	// use emitter vertex color channel
	MAP_HAIRCOLOR = -2,		// internal hair Base/Tip/bias color
	MAP_STRANDBASED = -1,	// natural cylindrical mapping, v goes along the strand
	MAP_NONE = 0,
	// 1..MAX_MAPCHANNELS-1 specify usage of emitter map channels
};


// strand cross-section types (Section radio button in the Parameters rollout)
enum CrossSectionType {
	CROSSSEC_ROUND,
	CROSSSEC_FLAT,
};


// a function which calculates the hair color based on the t parameter going
// from 0 at the base of the strand to 1 at the tip, and the Color parameters
// from the Shading rollout
inline Color GetBiasedCol( const Color &cBase, const Color &cTip, float t, float bias ) {
	t = t+bias;
	if ( t < 0 ) t = 0;
	if ( t > 1 ) t = 1;
	return (1-t) * cBase + t * cTip;
}


/*
// sample ShagExportData usage:
	Object			*obj;
	ShagExportData	*pShagExp;

	obj = pNode->EvalWorldState(t).obj;

	if ( !obj->CanConvertToType( SHAGEXPORTDATA_CLASS_ID ) ) ...	// not hair

	pShagExp = (ShagExportData*)obj->ConvertToType( t, SHAGEXPORTDATA_CLASS_ID );
	if ( !pShagExp ) ... // something's wrong

// NB: Don't delete the pShagExp pointer after finishing with it!
*/


class ShagExportData {
public:
	int			nStrands;
	int			nMapChans;		// the last allocated channel + 1, or 0 for no channels
	int			nKnots;			// number of bezier knots
	Point3		**pKnots;		// [0..nStrands-1][0..3*(nKnots-1)] bezier knots and handles in object space
	float		*pvCut;			// cut point of each strand, from 0=root to 1=tip
	float		*pvParamCut;	// cut point of the length-depending parameters (color, UV) when model cutters with parameter cutting are used

	Interval	validTopo;		// validity interval of the hair topology (# of strands and # of knots)

	float		thick;				// constant thickness value
	float		*pThickPerStrand;	// thickness per strand when randomization or thickmap is used
	float		*pThickCurve;		// variable thickness curve, sampled at knots
	float		**pThickPerKnot;	// thickness at knot level (used for combined variable thickness and model cutting)
	// to get the thickness at each strand knot use GetThickness() below

	int			crossSection;	// CROSSSEC_ROUND or CROSSSEC_FLAT
	Point3		*pOrient;		// orientations of the strands for CROSSSEC_FLAT in object space
	// to get the normal at each strand point (having the tangent) use:
	// Point3 normal = tangent ^ ( pOrient[iStrand] ^ tangent );

	// to get the uv coords at a given strand point use the GetUV() method below

	float		colbias;		// -1.0f .. 1.0f
	Color		colBase, colTip;// constant values (not emitter-derived and not randomized)
	Color		*pcolBase, *pcolTip;// per strand values when emitter-derived or randomized
	// we put here the emitter-derived values for the channels that use them
	UVVert		**pUVW;			// [0..nMapChans-1][0..nStrands-1]
	int			mapTypes[MAX_MAPCHANNELS];	// contains MapType values

	ShagExportData() { Init(); }

	~ShagExportData() { FreeAll(); }

	void Init() {
		nStrands = nMapChans = nKnots = 0;
		pKnots = NULL;
		pvCut = pvParamCut = NULL;
		validTopo.SetEmpty();
		thick = 0.0f;
		pThickPerStrand = pThickCurve = NULL;
		pThickPerKnot = NULL;
		crossSection = CROSSSEC_ROUND;
		pOrient = NULL;
		colbias = 0.0f;
		colBase.Black(); colTip.Black();
		pcolBase = pcolTip = NULL;
		pUVW = NULL;
		for ( int i=0; i < MAX_MAPCHANNELS; i++ ) mapTypes[i] = MAP_NONE;
	}

	void FreeAll() {
		if ( pKnots ) {
			for ( int i=0; i < nStrands; i++ )
				if ( pKnots[i] ) delete [] pKnots[i];
			delete [] pKnots;
		}
		if ( pvCut ) delete [] pvCut;
		if ( pvParamCut ) delete [] pvParamCut;
		if ( pThickPerStrand ) delete [] pThickPerStrand;
		if ( pThickCurve ) delete [] pThickCurve;
		if ( pThickPerKnot ) {
			for ( int i=0; i < nStrands; i++ )
				if ( pThickPerKnot[i] ) delete [] pThickPerKnot[i];
			delete [] pThickPerKnot;
		}
		if ( pOrient ) delete [] pOrient;
		if ( pcolBase ) delete [] pcolBase;
		if ( pcolTip ) delete [] pcolTip;
		if ( pUVW ) {
			for ( int i=0; i < nMapChans; i++ )
				if ( pUVW[i] ) delete [] pUVW[i];
			delete [] pUVW;
		}
		Init();
	}

	UVVert GetUV( int iStrand, int iChan, int iKnot ) {
		float v = (float)iKnot / (nKnots-1);
		Color c;

		if ( pvParamCut ) v *= pvParamCut[iStrand];
		switch( mapTypes[iChan] ) {
			case MAP_STRANDBASED:
				return UVVert( 0.5f, v, 0.0f );
			case MAP_HAIRCOLOR: {
				c = GetBiasedCol( pcolBase ? pcolBase[iStrand] : colBase,
						pcolTip ? pcolTip[iStrand] : colTip,
						v, colbias );
				return UVVert( c.r, c.g, c.b );
			}
			case MAP_NONE: return UVVert(0,0,0);
			default:
				return pUVW[iChan][iStrand];
		}
	}

	float GetThickness( int iStrand, int iKnot ) {
		if ( pThickPerKnot ) return pThickPerKnot[iStrand][iKnot];
		float thickness = pThickPerStrand ? pThickPerStrand[iStrand] : thick;
		if ( pThickCurve ) thickness *= pThickCurve[iKnot];
		return thickness;
	}

	int Size() {
		int sz = sizeof(ShagExportData);

		for ( int i=0; i < nStrands; i++ ) if ( pKnots[i] )
			sz += (3*nKnots-2) * sizeof(Point3);
		if ( pThickPerKnot )
			for ( i=0; i < nStrands; i++ ) if ( pThickPerKnot[i] )
				sz += nKnots * sizeof(float);
		if ( pThickPerStrand ) sz += nStrands * sizeof(pThickPerStrand[0]);
		if ( pThickCurve ) sz += nKnots * sizeof(pThickCurve[0]);
		if ( pOrient ) sz += nStrands * sizeof(Point3);
		if ( pcolBase ) sz += nStrands * sizeof(pcolBase[0]);
		if ( pcolTip ) sz += nStrands * sizeof(pcolTip[0]);
		if ( pUVW ) for ( i=0; i < nMapChans; i++ ) if ( pUVW[i] )
			sz += nStrands * sizeof(pUVW[i][0]);
		return sz;
	}
};


#endif		// _SHAGHAIREXPORT_H_
